import json
import warnings
from traceback import format_exc

import stripe
from django.db import models
from django.utils.functional import cached_property

from .. import settings as djstripe_settings
from ..context_managers import stripe_temporary_api_version
from ..fields import JSONField
from ..signals import webhook_processing_error
from ..utils import fix_django_headers
from .base import logger
from .core import Event


def _get_version():
    from .. import __version__

    return __version__


class WebhookEventTrigger(models.Model):
    """
    An instance of a request that reached the server endpoint for Stripe webhooks.

    Webhook Events are initially **UNTRUSTED**, as it is possible for any web entity to
    post any data to our webhook url. Data posted may be valid Stripe information,
    garbage, or even malicious.
    The 'valid' flag in this model monitors this.
    """

    id = models.BigAutoField(primary_key=True)
    remote_ip = models.GenericIPAddressField(
        help_text="IP address of the request client."
    )
    headers = JSONField()
    body = models.TextField(blank=True)
    valid = models.BooleanField(
        default=False,
        help_text="Whether or not the webhook event has passed validation",
    )
    processed = models.BooleanField(
        default=False,
        help_text="Whether or not the webhook event has been successfully processed",
    )
    exception = models.CharField(max_length=128, blank=True)
    traceback = models.TextField(
        blank=True, help_text="Traceback if an exception was thrown during processing"
    )
    event = models.ForeignKey(
        "Event",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        help_text="Event object contained in the (valid) Webhook",
    )
    djstripe_version = models.CharField(
        max_length=32,
        default=_get_version,  # Needs to be a callable, otherwise it's a db default.
        help_text="The version of dj-stripe when the webhook was received",
    )
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    @classmethod
    def from_request(cls, request):
        """
        Create, validate and process a WebhookEventTrigger given a Django
        request object.

        The process is three-fold:
        1. Create a WebhookEventTrigger object from a Django request.
        2. Validate the WebhookEventTrigger as a Stripe event using the API.
        3. If valid, process it into an Event object (and child resource).
        """

        headers = fix_django_headers(request.META)
        assert headers
        try:
            body = request.body.decode(request.encoding or "utf-8")
        except Exception:
            body = "(error decoding body)"

        ip = request.META.get("REMOTE_ADDR")
        if not ip:
            warnings.warn(
                "Could not determine remote IP (missing REMOTE_ADDR). "
                "This is likely an issue with your wsgi/server setup."
            )
            ip = "0.0.0.0"
        obj = cls.objects.create(headers=headers, body=body, remote_ip=ip)

        try:
            obj.valid = obj.validate()
            if obj.valid:
                if djstripe_settings.WEBHOOK_EVENT_CALLBACK:
                    # If WEBHOOK_EVENT_CALLBACK, pass it for processing
                    djstripe_settings.WEBHOOK_EVENT_CALLBACK(obj)
                else:
                    # Process the item (do not save it, it'll get saved below)
                    obj.process(save=False)
        except Exception as e:
            max_length = WebhookEventTrigger._meta.get_field("exception").max_length
            obj.exception = str(e)[:max_length]
            obj.traceback = format_exc()

            # Send the exception as the webhook_processing_error signal
            webhook_processing_error.send(
                sender=WebhookEventTrigger,
                exception=e,
                data=getattr(e, "http_body", ""),
            )

            # re-raise the exception so Django sees it
            raise e
        finally:
            obj.save()

        return obj

    @cached_property
    def json_body(self):
        try:
            return json.loads(self.body)
        except ValueError:
            return {}

    @property
    def is_test_event(self):
        event_id = self.json_body.get("id")
        return event_id and event_id.endswith("_00000000000000")

    def validate(self, api_key=None):
        """
        The original contents of the Event message must be confirmed by
        refetching it and comparing the fetched data with the original data.

        This function makes an API call to Stripe to redownload the Event data
        and returns whether or not it matches the WebhookEventTrigger data.
        """

        local_data = self.json_body
        if "id" not in local_data or "livemode" not in local_data:
            return False

        if self.is_test_event:
            logger.info("Test webhook received: {}".format(local_data))
            return False

        if djstripe_settings.WEBHOOK_VALIDATION is None:
            # validation disabled
            return True
        elif (
            djstripe_settings.WEBHOOK_VALIDATION == "verify_signature"
            and djstripe_settings.WEBHOOK_SECRET
        ):
            try:
                stripe.WebhookSignature.verify_header(
                    self.body,
                    self.headers.get("stripe-signature"),
                    djstripe_settings.WEBHOOK_SECRET,
                    djstripe_settings.WEBHOOK_TOLERANCE,
                )
            except stripe.error.SignatureVerificationError:
                return False
            else:
                return True

        livemode = local_data["livemode"]
        api_key = api_key or djstripe_settings.get_default_api_key(livemode)

        # Retrieve the event using the api_version specified in itself
        with stripe_temporary_api_version(local_data["api_version"], validate=False):
            remote_data = Event.stripe_class.retrieve(
                id=local_data["id"], api_key=api_key
            )

        return local_data["data"] == remote_data["data"]

    def process(self, save=True):
        # Reset traceback and exception in case of reprocessing
        self.exception = ""
        self.traceback = ""

        self.event = Event.process(self.json_body)
        self.processed = True
        if save:
            self.save()

        return self.event



"""
Utility functions used for syncing data.
"""
from stripe.error import InvalidRequestError

from .models import Customer


def sync_subscriber(subscriber):
    """Sync a Customer with Stripe api data."""
    customer, _created = Customer.get_or_create(subscriber=subscriber)
    try:
        customer.sync_from_stripe_data(customer.api_retrieve())
        customer._sync_subscriptions()
        customer._sync_invoices()
        customer._sync_cards()
        customer._sync_charges()
    except InvalidRequestError as e:
        print("ERROR: " + str(e))
    return customer



"""
dj-stripe - Django + Stripe Made Easy
"""
import pkg_resources
from django.apps import AppConfig

__version__ = pkg_resources.require("dj-stripe")[0].version

default_app_config = "djstripe.DjstripeAppConfig"


class DjstripeAppConfig(AppConfig):
    """
    An AppConfig for dj-stripe which loads system checks
    and event handlers once Django is ready.
    """

    name = "djstripe"

    def ready(self):
        import stripe
        from . import (  # noqa: Register the checks and event handlers
            checks,
            event_handlers,
        )

        # Set app info
        # https://stripe.com/docs/building-plugins#setappinfo
        stripe.set_app_info(
            "dj-stripe",
            version=__version__,
            url="https://github.com/dj-stripe/dj-stripe",
        )



# Generated by Django 2.2.5 on 2019-09-11 23:33

import django.db.models.deletion
from django.db import migrations, models

import djstripe.enums
import djstripe.fields


def fix_djstripepaymentmethod_index_name_forwards(apps, schema_editor):
    # Altering the index is required because while we changed the name of old PaymentMethod model to
    # DjStripePaymentMethod, the migrations didn't update the names of index.
    # In the current migration, we create a new PaymentMethod model, hence before creating it, its
    # better to rename the old index.
    if not schema_editor.connection.vendor.startswith("postgres"):
        return
    schema_editor.execute(
        "ALTER INDEX djstripe_paymentmethod_id_0b9251df_like rename TO djstripe_paymentmethod_legacy_id_0b9251df_like"
    )


def fix_djstripepaymentmethod_index_name_backwards(apps, schema_editor):
    if not schema_editor.connection.vendor.startswith("postgres"):
        return
    schema_editor.execute(
        "ALTER INDEX djstripe_paymentmethod_legacy_id_0b9251df_like rename TO djstripe_paymentmethod_id_0b9251df_like"
    )


class Migration(migrations.Migration):

    dependencies = [
        ("djstripe", "0003_auto_20181117_2328_squashed_0004_auto_20190227_2114")
    ]

    operations = [
        # Altering the index is required because while we have changed the name of old PaymentMethod model to
        # DjStripePaymentMethod, the migrations didn't update the names of index.
        # In the current migration, we create a new PaymentMethod model, hence before creating it, its
        # better to rename the old index.
        migrations.RunPython(
            code=fix_djstripepaymentmethod_index_name_forwards,
            reverse_code=fix_djstripepaymentmethod_index_name_backwards,
        ),
        migrations.AlterField(
            model_name="subscriptionitem",
            name="quantity",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="The quantity of the plan to which the customer should be subscribed.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="invoice",
            name="auto_advance",
            field=models.NullBooleanField(
                help_text="Controls whether Stripe will perform automatic collection of the invoice. When false, the invoice’s state will not automatically advance without an explicit action."
            ),
        ),
        migrations.RenameField(
            model_name="account", old_name="business_logo", new_name="branding_icon"
        ),
        migrations.AddField(
            model_name="account",
            name="branding_logo",
            field=models.ForeignKey(
                help_text="A logo for the account that will be used in Checkout instead of the icon and without the account’s name next to it if provided. Must be at least 128px x 128px.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="logo_account",
                to="djstripe.FileUpload",
            ),
        ),
        migrations.AddField(
            model_name="account",
            name="business_profile",
            field=djstripe.fields.JSONField(
                blank=True,
                help_text="Optional information related to the business.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="account",
            name="business_type",
            field=djstripe.fields.StripeEnumField(
                blank=True,
                default="",
                enum=djstripe.enums.BusinessType,
                help_text="The business type.",
                max_length=10,
            ),
        ),
        migrations.AddField(
            model_name="account",
            name="company",
            field=djstripe.fields.JSONField(
                blank=True,
                help_text="Information about the company or business. This field is null unless business_type is set to company.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="account",
            name="individual",
            field=djstripe.fields.JSONField(
                blank=True,
                help_text="Information about the person represented by the account. This field is null unless business_type is set to individual.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="account",
            name="requirements",
            field=djstripe.fields.JSONField(
                blank=True,
                help_text="Information about the requirements for the account, including what information needs to be collected, and by when.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="account",
            name="settings",
            field=djstripe.fields.JSONField(
                blank=True,
                help_text="Account options for customizing how the account functions within Stripe.",
                null=True,
            ),
        ),
        migrations.AlterModelOptions(
            name="invoice", options={"ordering": ["-created"]}
        ),
        migrations.RenameField(
            model_name="invoice",
            old_name="application_fee",
            new_name="application_fee_amount",
        ),
        migrations.RemoveField(model_name="invoice", name="date"),
        migrations.AddField(
            model_name="invoice",
            name="status_transitions",
            field=djstripe.fields.JSONField(blank=True, null=True),
        ),
        migrations.RenameField(
            model_name="customer", old_name="account_balance", new_name="balance"
        ),
        migrations.CreateModel(
            name="PaymentMethod",
            fields=[
                (
                    "djstripe_id",
                    models.BigAutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("id", djstripe.fields.StripeIdField(max_length=255, unique=True)),
                (
                    "livemode",
                    models.NullBooleanField(
                        default=None,
                        help_text="Null here indicates that the livemode status is unknown or was previously unrecorded. Otherwise, this field indicates whether this record comes from Stripe test mode or live mode operation.",
                    ),
                ),
                (
                    "created",
                    djstripe.fields.StripeDateTimeField(
                        blank=True,
                        help_text="The datetime this object was created in stripe.",
                        null=True,
                    ),
                ),
                (
                    "metadata",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="A set of key/value pairs that you can attach to an object. It can be useful for storing additional information about an object in a structured format.",
                        null=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="A description of this object.", null=True
                    ),
                ),
                ("djstripe_created", models.DateTimeField(auto_now_add=True)),
                ("djstripe_updated", models.DateTimeField(auto_now=True)),
                (
                    "billing_details",
                    djstripe.fields.JSONField(
                        help_text="Billing information associated with the PaymentMethod that may be used or required by particular types of payment methods."
                    ),
                ),
                (
                    "card",
                    djstripe.fields.JSONField(
                        help_text="If this is a card PaymentMethod, this hash contains details about the card."
                    ),
                ),
                (
                    "card_present",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="If this is an card_present PaymentMethod, this hash contains details about the Card Present payment method.",
                        null=True,
                    ),
                ),
                (
                    "type",
                    models.CharField(
                        blank=True,
                        help_text="The type of the PaymentMethod. An additional hash is included on the PaymentMethod with a name matching this value. It contains additional information specific to the PaymentMethod type.",
                        max_length=255,
                    ),
                ),
                (
                    "customer",
                    models.ForeignKey(
                        blank=True,
                        help_text="Customer to which this PaymentMethod is saved.This will not be set when the PaymentMethod has not been saved to a Customer.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="payment_methods",
                        to="djstripe.Customer",
                    ),
                ),
            ],
            options={"get_latest_by": "created", "abstract": False},
        ),
        migrations.AlterField(
            model_name="payout",
            name="destination",
            field=models.ForeignKey(
                help_text="Bank account or card the payout was sent to.",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="djstripe.BankAccount",
            ),
        ),
        migrations.CreateModel(
            name="SetupIntent",
            fields=[
                (
                    "djstripe_id",
                    models.BigAutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("id", djstripe.fields.StripeIdField(max_length=255, unique=True)),
                (
                    "livemode",
                    models.NullBooleanField(
                        default=None,
                        help_text="Null here indicates that the livemode status is unknown or was previously unrecorded. Otherwise, this field indicates whether this record comes from Stripe test mode or live mode operation.",
                    ),
                ),
                (
                    "created",
                    djstripe.fields.StripeDateTimeField(
                        blank=True,
                        help_text="The datetime this object was created in stripe.",
                        null=True,
                    ),
                ),
                (
                    "metadata",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="A set of key/value pairs that you can attach to an object. It can be useful for storing additional information about an object in a structured format.",
                        null=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="A description of this object.", null=True
                    ),
                ),
                ("djstripe_created", models.DateTimeField(auto_now_add=True)),
                ("djstripe_updated", models.DateTimeField(auto_now=True)),
                (
                    "application",
                    models.CharField(
                        blank=True,
                        help_text="ID of the Connect application that created the SetupIntent.",
                        max_length=255,
                    ),
                ),
                (
                    "cancellation_reason",
                    djstripe.fields.StripeEnumField(
                        blank=True,
                        enum=djstripe.enums.SetupIntentCancellationReason,
                        help_text="Reason for cancellation of this SetupIntent, one of abandoned, requested_by_customer, or duplicate",
                        max_length=21,
                    ),
                ),
                (
                    "client_secret",
                    models.TextField(
                        blank=True,
                        help_text="The client secret of this SetupIntent. Used for client-side retrieval using a publishable key.",
                        max_length=5000,
                    ),
                ),
                (
                    "last_setup_error",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="The error encountered in the previous SetupIntent confirmation.",
                        null=True,
                    ),
                ),
                (
                    "next_action",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="If present, this property tells you what actions you need to take inorder for your customer to continue payment setup.",
                        null=True,
                    ),
                ),
                (
                    "payment_method_types",
                    djstripe.fields.JSONField(
                        help_text="The list of payment method types (e.g. card) that this PaymentIntent is allowed to use."
                    ),
                ),
                (
                    "status",
                    djstripe.fields.StripeEnumField(
                        enum=djstripe.enums.SetupIntentStatus,
                        help_text="Status of this SetupIntent, one of requires_payment_method, requires_confirmation, requires_action, processing, canceled, or succeeded.",
                        max_length=23,
                    ),
                ),
                (
                    "usage",
                    djstripe.fields.StripeEnumField(
                        default="off_session",
                        enum=djstripe.enums.IntentUsage,
                        help_text="Indicates how the payment method is intended to be used in the future.",
                        max_length=11,
                    ),
                ),
                (
                    "customer",
                    models.ForeignKey(
                        blank=True,
                        help_text="Customer this SetupIntent belongs to, if one exists.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="djstripe.Customer",
                    ),
                ),
                (
                    "on_behalf_of",
                    models.ForeignKey(
                        blank=True,
                        help_text="The account (if any) for which the setup is intended.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="djstripe.Account",
                    ),
                ),
                (
                    "payment_method",
                    models.ForeignKey(
                        blank=True,
                        help_text="Payment method used in this PaymentIntent.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="djstripe.PaymentMethod",
                    ),
                ),
            ],
            options={"get_latest_by": "created", "abstract": False},
        ),
        migrations.CreateModel(
            name="PaymentIntent",
            fields=[
                (
                    "djstripe_id",
                    models.BigAutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("id", djstripe.fields.StripeIdField(max_length=255, unique=True)),
                (
                    "livemode",
                    models.NullBooleanField(
                        default=None,
                        help_text="Null here indicates that the livemode status is unknown or was previously unrecorded. Otherwise, this field indicates whether this record comes from Stripe test mode or live mode operation.",
                    ),
                ),
                (
                    "created",
                    djstripe.fields.StripeDateTimeField(
                        blank=True,
                        help_text="The datetime this object was created in stripe.",
                        null=True,
                    ),
                ),
                (
                    "metadata",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="A set of key/value pairs that you can attach to an object. It can be useful for storing additional information about an object in a structured format.",
                        null=True,
                    ),
                ),
                ("djstripe_created", models.DateTimeField(auto_now_add=True)),
                ("djstripe_updated", models.DateTimeField(auto_now=True)),
                (
                    "amount",
                    djstripe.fields.StripeQuantumCurrencyAmountField(
                        help_text="Amount (in cents) intended to be collected by this PaymentIntent."
                    ),
                ),
                (
                    "amount_capturable",
                    djstripe.fields.StripeQuantumCurrencyAmountField(
                        help_text="Amount (in cents) that can be captured from this PaymentIntent."
                    ),
                ),
                (
                    "amount_received",
                    djstripe.fields.StripeQuantumCurrencyAmountField(
                        help_text="Amount (in cents) that was collected by this PaymentIntent."
                    ),
                ),
                (
                    "canceled_at",
                    djstripe.fields.StripeDateTimeField(
                        blank=True,
                        default=None,
                        help_text="Populated when status is canceled, this is the time at which the PaymentIntent was canceled. Measured in seconds since the Unix epoch.",
                        null=True,
                    ),
                ),
                (
                    "cancellation_reason",
                    djstripe.fields.StripeEnumField(
                        blank=True,
                        enum=djstripe.enums.PaymentIntentCancellationReason,
                        help_text="Reason for cancellation of this PaymentIntent, either user-provided (duplicate, fraudulent, requested_by_customer, or abandoned) or generated by Stripe internally (failed_invoice, void_invoice, or automatic).",
                        max_length=21,
                    ),
                ),
                (
                    "capture_method",
                    djstripe.fields.StripeEnumField(
                        enum=djstripe.enums.CaptureMethod,
                        help_text="Capture method of this PaymentIntent, one of automatic or manual.",
                        max_length=9,
                    ),
                ),
                (
                    "client_secret",
                    models.TextField(
                        help_text="The client secret of this PaymentIntent. Used for client-side retrieval using a publishable key.",
                        max_length=5000,
                    ),
                ),
                (
                    "confirmation_method",
                    djstripe.fields.StripeEnumField(
                        enum=djstripe.enums.ConfirmationMethod,
                        help_text="Confirmation method of this PaymentIntent, one of manual or automatic.",
                        max_length=9,
                    ),
                ),
                (
                    "currency",
                    djstripe.fields.StripeCurrencyCodeField(
                        help_text="Three-letter ISO currency code", max_length=3
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        default="",
                        help_text="An arbitrary string attached to the object. Often useful for displaying to users.",
                        max_length=1000,
                    ),
                ),
                (
                    "last_payment_error",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="The payment error encountered in the previous PaymentIntent confirmation.",
                        null=True,
                    ),
                ),
                (
                    "next_action",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="If present, this property tells you what actions you need to take in order for your customer to fulfill a payment using the provided source.",
                        null=True,
                    ),
                ),
                (
                    "payment_method_types",
                    djstripe.fields.JSONField(
                        help_text="The list of payment method types (e.g. card) that this PaymentIntent is allowed to use."
                    ),
                ),
                (
                    "receipt_email",
                    models.CharField(
                        blank=True,
                        help_text="Email address that the receipt for the resulting payment will be sent to.",
                        max_length=255,
                    ),
                ),
                (
                    "setup_future_usage",
                    djstripe.fields.StripeEnumField(
                        blank=True,
                        enum=djstripe.enums.IntentUsage,
                        help_text="Indicates that you intend to make future payments with this PaymentIntent’s payment method. If present, the payment method used with this PaymentIntent can be attached to a Customer, even after the transaction completes. Use `on_session` if you intend to only reuse the payment method when your customer is present in your checkout flow. Use `off_session` if your customer may or may not be in your checkout flow. Stripe uses `setup_future_usage` to dynamically optimize your payment flow and comply with regional legislation and network rules. For example, if your customer is impacted by SCA, using `off_session` will ensure that they are authenticated while processing this PaymentIntent. You will then be able to make later off-session payments for this customer.",
                        max_length=11,
                        null=True,
                    ),
                ),
                (
                    "shipping",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="Shipping information for this PaymentIntent.",
                        null=True,
                    ),
                ),
                (
                    "statement_descriptor",
                    models.CharField(
                        blank=True,
                        help_text="For non-card charges, you can use this value as the complete description that appears on your customers’ statements. Must contain at least one letter, maximum 22 characters.",
                        max_length=22,
                    ),
                ),
                (
                    "status",
                    djstripe.fields.StripeEnumField(
                        enum=djstripe.enums.PaymentIntentStatus,
                        help_text="Status of this PaymentIntent, one of requires_payment_method, requires_confirmation, requires_action, processing, requires_capture, canceled, or succeeded. You can read more about PaymentIntent statuses here.",
                        max_length=23,
                    ),
                ),
                (
                    "transfer_data",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="The data with which to automatically create a Transfer when the payment is finalized. See the PaymentIntents Connect usage guide for details.",
                        null=True,
                    ),
                ),
                (
                    "transfer_group",
                    models.CharField(
                        blank=True,
                        help_text="A string that identifies the resulting payment as part of a group. See the PaymentIntents Connect usage guide for details.",
                        max_length=255,
                    ),
                ),
                (
                    "customer",
                    models.ForeignKey(
                        help_text="Customer this PaymentIntent is for if one exists.",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djstripe.Customer",
                    ),
                ),
                (
                    "on_behalf_of",
                    models.ForeignKey(
                        blank=True,
                        help_text="The account (if any) for which the funds of the PaymentIntent are intended.",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djstripe.Account",
                    ),
                ),
                (
                    "payment_method",
                    models.ForeignKey(
                        blank=True,
                        help_text="Payment method used in this PaymentIntent.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="djstripe.PaymentMethod",
                    ),
                ),
            ],
            options={"get_latest_by": "created", "abstract": False},
        ),
        migrations.AddField(
            model_name="charge",
            name="payment_intent",
            field=models.ForeignKey(
                help_text="PaymentIntent associated with this charge, if one exists.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="charges",
                to="djstripe.PaymentIntent",
            ),
        ),
        migrations.AddField(
            model_name="invoice",
            name="payment_intent",
            field=models.OneToOneField(
                help_text="The PaymentIntent associated with this invoice. The PaymentIntent is generated when the invoice is finalized, and can then be used to pay the invoice.Note that voiding an invoice will cancel the PaymentIntent",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="djstripe.PaymentIntent",
            ),
        ),
        migrations.AddField(
            model_name="subscription",
            name="pending_setup_intent",
            field=models.ForeignKey(
                blank=True,
                help_text="We can use this SetupIntent to collect user authentication when creating a subscription without immediate payment or updating a subscription’s payment method, allowing you to optimize for off-session payments.",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="setup_intents",
                to="djstripe.SetupIntent",
            ),
        ),
        migrations.AddField(
            model_name="customer",
            name="address",
            field=djstripe.fields.JSONField(
                blank=True, help_text="The customer's address.", null=True
            ),
        ),
        migrations.AddField(
            model_name="customer",
            name="invoice_prefix",
            field=models.CharField(
                blank=True,
                default="",
                help_text="The prefix for the customer used to generate unique invoice numbers.",
                max_length=255,
            ),
        ),
        migrations.AddField(
            model_name="customer",
            name="invoice_settings",
            field=djstripe.fields.JSONField(
                blank=True,
                help_text="The customer's default invoice settings.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="customer",
            name="name",
            field=models.TextField(
                blank=True,
                default="",
                help_text="The customer's full name or business name.",
                max_length=5000,
            ),
        ),
        migrations.AddField(
            model_name="customer",
            name="phone",
            field=models.TextField(
                blank=True,
                default="",
                help_text="The customer's phone number.",
                max_length=5000,
            ),
        ),
        migrations.AddField(
            model_name="customer",
            name="preferred_locales",
            field=djstripe.fields.JSONField(
                blank=True,
                help_text="The customer's preferred locales (languages), ordered by preference.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="customer",
            name="tax_exempt",
            field=djstripe.fields.StripeEnumField(
                default="",
                enum=djstripe.enums.CustomerTaxExempt,
                help_text='Describes the customer\'s tax exemption status. When set to reverse, invoice and receipt PDFs include the text "Reverse charge".',
                max_length=7,
            ),
        ),
        migrations.AlterField(
            model_name="bankaccount",
            name="account",
            field=models.ForeignKey(
                blank=True,
                help_text="The account the charge was made on behalf of. Null here indicates that this value was never set.",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="bank_account",
                to="djstripe.Account",
            ),
        ),
        migrations.AlterField(
            model_name="charge",
            name="fraud_details",
            field=djstripe.fields.JSONField(
                blank=True,
                help_text="Hash with information on fraud assessments for the charge.",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="charge",
            name="outcome",
            field=djstripe.fields.JSONField(
                blank=True,
                help_text="Details about whether or not the payment was accepted, and why.",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="status",
            field=djstripe.fields.StripeEnumField(
                enum=djstripe.enums.SubscriptionStatus,
                help_text="The status of this subscription.",
                max_length=18,
            ),
        ),
        migrations.AddField(
            model_name="customer",
            name="default_payment_method",
            field=models.ForeignKey(
                blank=True,
                help_text="default payment method used for subscriptions and invoices for the customer.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="djstripe.PaymentMethod",
            ),
        ),
        migrations.AlterField(
            model_name="balancetransaction",
            name="type",
            field=djstripe.fields.StripeEnumField(
                enum=djstripe.enums.BalanceTransactionType, max_length=29
            ),
        ),
        migrations.AddField(
            model_name="charge",
            name="payment_method_details",
            field=djstripe.fields.JSONField(
                blank=True,
                help_text="Details about the payment method at the time of the transaction.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="charge",
            name="receipt_url",
            field=models.TextField(
                blank=True,
                default="",
                help_text="This is the URL to view the receipt for this charge. The receipt is kept up-to-date to the latest state of the charge, including any refunds. If the charge is for an Invoice, the receipt will be stylized as an Invoice receipt.",
                max_length=5000,
            ),
        ),
        migrations.CreateModel(
            name="Session",
            fields=[
                (
                    "djstripe_id",
                    models.BigAutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("id", djstripe.fields.StripeIdField(max_length=255, unique=True)),
                (
                    "livemode",
                    models.NullBooleanField(
                        default=None,
                        help_text="Null here indicates that the livemode status is unknown or was previously unrecorded. Otherwise, this field indicates whether this record comes from Stripe test mode or live mode operation.",
                    ),
                ),
                (
                    "created",
                    djstripe.fields.StripeDateTimeField(
                        blank=True,
                        help_text="The datetime this object was created in stripe.",
                        null=True,
                    ),
                ),
                (
                    "metadata",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="A set of key/value pairs that you can attach to an object. It can be useful for storing additional information about an object in a structured format.",
                        null=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="A description of this object.", null=True
                    ),
                ),
                ("djstripe_created", models.DateTimeField(auto_now_add=True)),
                ("djstripe_updated", models.DateTimeField(auto_now=True)),
                (
                    "billing_address_collection",
                    djstripe.fields.StripeEnumField(
                        blank=True,
                        enum=djstripe.enums.SessionBillingAddressCollection,
                        help_text="The value (auto or required) for whether Checkoutcollected the customer’s billing address.",
                        max_length=8,
                    ),
                ),
                (
                    "cancel_url",
                    models.TextField(
                        blank=True,
                        help_text="The URL the customer will be directed to if theydecide to cancel payment and return to your website.",
                        max_length=5000,
                    ),
                ),
                (
                    "client_reference_id",
                    models.TextField(
                        blank=True,
                        help_text="A unique string to reference the Checkout Session.This can be a customer ID, a cart ID, or similar, andcan be used to reconcile the session with your internal systems.",
                        max_length=5000,
                    ),
                ),
                (
                    "customer_email",
                    models.CharField(
                        blank=True,
                        help_text="If provided, this value will be used when the Customer object is created.",
                        max_length=255,
                    ),
                ),
                (
                    "display_items",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="The line items, plans, or SKUs purchased by the customer.",
                        null=True,
                    ),
                ),
                (
                    "locale",
                    models.CharField(
                        blank=True,
                        help_text="The IETF language tag of the locale Checkout is displayed in.If blank or auto, the browser’s locale is used.",
                        max_length=255,
                    ),
                ),
                (
                    "payment_method_types",
                    djstripe.fields.JSONField(
                        help_text="The list of payment method types (e.g. card) that this Checkout Session is allowed to accept."
                    ),
                ),
                (
                    "submit_type",
                    djstripe.fields.StripeEnumField(
                        blank=True,
                        enum=djstripe.enums.SubmitTypeStatus,
                        help_text="Describes the type of transaction being performed by Checkoutin order to customize relevant text on the page, such as the submit button.",
                        max_length=6,
                    ),
                ),
                (
                    "success_url",
                    models.TextField(
                        blank=True,
                        help_text="The URL the customer will be directed to after the payment or subscriptioncreation is successful.",
                        max_length=5000,
                    ),
                ),
                (
                    "customer",
                    models.ForeignKey(
                        help_text="Customer this Checkout is for if one exists.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="djstripe.Customer",
                    ),
                ),
                (
                    "payment_intent",
                    models.ForeignKey(
                        help_text="PaymentIntent created if SKUs or line items were provided.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="djstripe.PaymentIntent",
                    ),
                ),
                (
                    "subscription",
                    models.ForeignKey(
                        help_text="Subscription created if one or more plans were provided.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="djstripe.Subscription",
                    ),
                ),
                (
                    "mode",
                    djstripe.fields.StripeEnumField(
                        blank=True,
                        enum=djstripe.enums.SessionMode,
                        help_text="The mode of the Checkout Session, one of payment, setup, or subscription.",
                        max_length=12,
                    ),
                ),
            ],
            options={"get_latest_by": "created", "abstract": False},
        ),
        migrations.AddField(
            model_name="charge",
            name="payment_method",
            field=models.ForeignKey(
                help_text="PaymentMethod used in this charge.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="charges",
                to="djstripe.PaymentMethod",
            ),
        ),
    ]



"""
Urls related to the djstripe app.

Wire this into the root URLConf this way::

    path("stripe/", include("djstripe.urls", namespace="djstripe")),
    # url can be changed
    # Call to 'djstripe.urls' and 'namespace' must stay as is

Call it from reverse()::

    reverse("djstripe:subscribe")

Call from url tag::

    {% url "djstripe:subscribe" %}
"""
from django.urls import re_path

from . import settings as app_settings
from . import views

app_name = "djstripe"

urlpatterns = [
    # Webhook
    re_path(
        app_settings.DJSTRIPE_WEBHOOK_URL,
        views.ProcessWebhookView.as_view(),
        name="webhook",
    )
]



import operator
from collections import OrderedDict

from django.utils.translation import gettext_lazy as _


class EnumMetaClass(type):
    @classmethod
    def __prepare__(self, name, bases):
        return OrderedDict()

    def __new__(self, name, bases, classdict):
        members = []
        keys = {}
        choices = OrderedDict()
        for key, value in classdict.items():
            if key.startswith("__"):
                continue
            members.append(key)
            if isinstance(value, tuple):
                value, alias = value
                keys[alias] = key
            else:
                alias = None
            keys[alias or key] = key
            choices[alias or key] = value

        for k, v in keys.items():
            classdict[v] = k

        classdict["__choices__"] = choices
        classdict["__members__"] = members

        # Note: Differences between Python 2.x and Python 3.x force us to
        # explicitly use unicode here, and to explicitly sort the list. In
        # Python 2.x, class members are unordered and so the ordering will
        # vary on different systems based on internal hashing. Without this
        # Django will continually require new no-op migrations.
        classdict["choices"] = tuple(
            (str(k), str(v))
            for k, v in sorted(choices.items(), key=operator.itemgetter(0))
        )

        return type.__new__(self, name, bases, classdict)


class Enum(metaclass=EnumMetaClass):
    pass


class ApiErrorCode(Enum):
    """
    Charge failure error codes.

    https://stripe.com/docs/error-codes
    """

    account_already_exists = _("Account already exists")
    account_country_invalid_address = _("Account country invalid address")
    account_invalid = _("Account invalid")
    account_number_invalid = _("Account number invalid")
    alipay_upgrade_required = _("Alipay upgrade required")
    amount_too_large = _("Amount too large")
    amount_too_small = _("Amount too small")
    api_key_expired = _("Api key expired")
    balance_insufficient = _("Balance insufficient")
    bank_account_exists = _("Bank account exists")
    bank_account_unusable = _("Bank account unusable")
    bank_account_unverified = _("Bank account unverified")
    bitcoin_upgrade_required = _("Bitcoin upgrade required")
    card_declined = _("Card was declined")
    charge_already_captured = _("Charge already captured")
    charge_already_refunded = _("Charge already refunded")
    charge_disputed = _("Charge disputed")
    charge_exceeds_source_limit = _("Charge exceeds source limit")
    charge_expired_for_capture = _("Charge expired for capture")
    country_unsupported = _("Country unsupported")
    coupon_expired = _("Coupon expired")
    customer_max_subscriptions = _("Customer max subscriptions")
    email_invalid = _("Email invalid")
    expired_card = _("Expired card")
    idempotency_key_in_use = _("Idempotency key in use")
    incorrect_address = _("Incorrect address")
    incorrect_cvc = _("Incorrect security code")
    incorrect_number = _("Incorrect number")
    incorrect_zip = _("ZIP code failed validation")
    instant_payouts_unsupported = _("Instant payouts unsupported")
    invalid_card_type = _("Invalid card type")
    invalid_charge_amount = _("Invalid charge amount")
    invalid_cvc = _("Invalid security code")
    invalid_expiry_month = _("Invalid expiration month")
    invalid_expiry_year = _("Invalid expiration year")
    invalid_number = _("Invalid number")
    invalid_source_usage = _("Invalid source usage")
    invoice_no_customer_line_items = _("Invoice no customer line items")
    invoice_no_subscription_line_items = _("Invoice no subscription line items")
    invoice_not_editable = _("Invoice not editable")
    invoice_upcoming_none = _("Invoice upcoming none")
    livemode_mismatch = _("Livemode mismatch")
    missing = _("No card being charged")
    not_allowed_on_standard_account = _("Not allowed on standard account")
    order_creation_failed = _("Order creation failed")
    order_required_settings = _("Order required settings")
    order_status_invalid = _("Order status invalid")
    order_upstream_timeout = _("Order upstream timeout")
    out_of_inventory = _("Out of inventory")
    parameter_invalid_empty = _("Parameter invalid empty")
    parameter_invalid_integer = _("Parameter invalid integer")
    parameter_invalid_string_blank = _("Parameter invalid string blank")
    parameter_invalid_string_empty = _("Parameter invalid string empty")
    parameter_missing = _("Parameter missing")
    parameter_unknown = _("Parameter unknown")
    parameters_exclusive = _("Parameters exclusive")
    payment_intent_authentication_failure = _("Payment intent authentication failure")
    payment_intent_incompatible_payment_method = _(
        "Payment intent incompatible payment method"
    )
    payment_intent_invalid_parameter = _("Payment intent invalid parameter")
    payment_intent_payment_attempt_failed = _("Payment intent payment attempt failed")
    payment_intent_unexpected_state = _("Payment intent unexpected state")
    payment_method_unactivated = _("Payment method unactivated")
    payment_method_unexpected_state = _("Payment method unexpected state")
    payouts_not_allowed = _("Payouts not allowed")
    platform_api_key_expired = _("Platform api key expired")
    postal_code_invalid = _("Postal code invalid")
    processing_error = _("Processing error")
    product_inactive = _("Product inactive")
    rate_limit = _("Rate limit")
    resource_already_exists = _("Resource already exists")
    resource_missing = _("Resource missing")
    routing_number_invalid = _("Routing number invalid")
    secret_key_required = _("Secret key required")
    sepa_unsupported_account = _("SEPA unsupported account")
    shipping_calculation_failed = _("Shipping calculation failed")
    sku_inactive = _("SKU inactive")
    state_unsupported = _("State unsupported")
    tax_id_invalid = _("Tax id invalid")
    taxes_calculation_failed = _("Taxes calculation failed")
    testmode_charges_only = _("Testmode charges only")
    tls_version_unsupported = _("TLS version unsupported")
    token_already_used = _("Token already used")
    token_in_use = _("Token in use")
    transfers_not_allowed = _("Transfers not allowed")
    upstream_order_creation_failed = _("Upstream order creation failed")
    url_invalid = _("URL invalid")

    # deprecated
    invalid_swipe_data = _("Invalid swipe data")


class AccountType(Enum):
    standard = _("Standard")
    express = _("Express")
    custom = _("Custom")


class BalanceTransactionStatus(Enum):
    available = _("Available")
    pending = _("Pending")


class BalanceTransactionType(Enum):
    adjustment = _("Adjustment")
    advance = _("Advance")
    advance_funding = _("Advance funding")
    application_fee = _("Application fee")
    application_fee_refund = _("Application fee refund")
    charge = _("Charge")
    connect_collection_transfer = _("Connect collection transfer")
    issuing_authorization_hold = _("Issuing authorization hold")
    issuing_authorization_release = _("Issuing authorization release")
    issuing_transaction = _("Issuing transaction")
    network_cost = _("Network cost")
    payment = _("Payment")
    payment_failure_refund = _("Payment failure refund")
    payment_refund = _("Payment refund")
    payout = _("Payout")
    payout_cancel = _("Payout cancellation")
    payout_failure = _("Payout failure")
    refund = _("Refund")
    refund_failure = _("Refund failure")
    reserve_transaction = _("Reserve transaction")
    reserved_funds = _("Reserved funds")
    stripe_fee = _("Stripe fee")
    stripe_fx_fee = _("Stripe fx fee")
    tax_fee = _("Tax fee")
    topup = _("Topup")
    topup_reversal = _("Topup reversal")
    transfer = _("Transfer")
    transfer_cancel = _("Transfer cancel")
    transfer_refund = _("Transfer refund")
    validation = _("Validation")


class BankAccountHolderType(Enum):
    individual = _("Individual")
    company = _("Company")


class BankAccountStatus(Enum):
    new = _("New")
    validated = _("Validated")
    verified = _("Verified")
    verification_failed = _("Verification failed")
    errored = _("Errored")


class BusinessType(Enum):
    individual = _("Individual")
    company = _("Company")


class CaptureMethod(Enum):
    automatic = _("Automatic")
    manual = _("Manual")


class CardCheckResult(Enum):
    pass_ = (_("Pass"), "pass")
    fail = _("Fail")
    unavailable = _("Unavailable")
    unchecked = _("Unchecked")


class CardBrand(Enum):
    AmericanExpress = (_("American Express"), "American Express")
    DinersClub = (_("Diners Club"), "Diners Club")
    Discover = _("Discover")
    JCB = _("JCB")
    MasterCard = _("MasterCard")
    UnionPay = _("UnionPay")
    Visa = _("Visa")
    Unknown = _("Unknown")


class CardFundingType(Enum):
    credit = _("Credit")
    debit = _("Debit")
    prepaid = _("Prepaid")
    unknown = _("Unknown")


class CardTokenizationMethod(Enum):
    apple_pay = _("Apple Pay")
    android_pay = _("Android Pay")


class ChargeStatus(Enum):
    succeeded = _("Succeeded")
    pending = _("Pending")
    failed = _("Failed")


class ConfirmationMethod(Enum):
    automatic = _("Automatic")
    manual = _("Manual")


class CouponDuration(Enum):
    once = _("Once")
    repeating = _("Multi-month")
    forever = _("Forever")


class CustomerTaxExempt(Enum):
    none = _("None")
    exempt = _("Exempt")
    reverse = _("Reverse")


class DisputeReason(Enum):
    duplicate = _("Duplicate")
    fraudulent = _("Fraudulent")
    subscription_canceled = _("Subscription canceled")
    product_unacceptable = _("Product unacceptable")
    product_not_received = _("Product not received")
    unrecognized = _("Unrecognized")
    credit_not_processed = _("Credit not processed")
    general = _("General")
    incorrect_account_details = _("Incorrect account details")
    insufficient_funds = _("Insufficient funds")
    bank_cannot_process = _("Bank cannot process")
    debit_not_authorized = _("Debit not authorized")
    customer_initiated = _("Customer-initiated")


class DisputeStatus(Enum):
    warning_needs_response = _("Warning needs response")
    warning_under_review = _("Warning under review")
    warning_closed = _("Warning closed")
    needs_response = _("Needs response")
    under_review = _("Under review")
    charge_refunded = _("Charge refunded")
    won = _("Won")
    lost = _("Lost")


class FileUploadPurpose(Enum):
    dispute_evidence = _("Dispute evidence")
    identity_document = _("Identity document")
    tax_document_user_upload = _("Tax document user upload")


class FileUploadType(Enum):
    pdf = _("PDF")
    jpg = _("JPG")
    png = _("PNG")
    csv = _("CSV")
    xls = _("XLS")
    xlsx = _("XLSX")
    docx = _("DOCX")


class InvoiceBillingReason(Enum):
    subscription_cycle = _("Subscription cycle")
    subscription_create = _("Subscription create")
    subscription_update = _("Subscription update")
    subscription = _("Subscription")
    manual = _("Manual")
    upcoming = _("Upcoming")
    subscription_threshold = _("Subscription threshold")


class InvoiceCollectionMethod(Enum):
    charge_automatically = _("Charge automatically")
    send_invoice = _("Send invoice")


# deprecated, will be removed in 2.3
InvoiceBilling = InvoiceCollectionMethod


class IntentUsage(Enum):
    on_session = _("On session")
    off_session = _("Off session")


class IntentStatus(Enum):
    """
    Status of Intents which apply both to PaymentIntents
    and SetupIntents.
    """

    requires_payment_method = _(
        "Intent created and requires a Payment Method to be attached."
    )
    requires_confirmation = _("Intent is ready to be confirmed.")
    requires_action = _("Payment Method require additional action, such as 3D secure.")
    processing = _("Required actions have been handled.")
    canceled = _(
        "Cancellation invalidates the intent for future confirmation and "
        "cannot be undone."
    )


# TODO - maybe refactor Enum so that inheritance works,
#  then PaymentIntentStatus/SetupIntentStatus can inherit from IntentStatus
class PaymentIntentStatus(Enum):
    requires_payment_method = _(
        "Intent created and requires a Payment Method to be attached."
    )
    requires_confirmation = _("Intent is ready to be confirmed.")
    requires_action = _("Payment Method require additional action, such as 3D secure.")
    processing = _("Required actions have been handled.")
    requires_capture = _("Capture the funds on the cards which have been put on holds.")
    canceled = _(
        "Cancellation invalidates the intent for future confirmation and "
        "cannot be undone."
    )
    succeeded = _("The funds are in your account.")


class SetupIntentStatus(Enum):
    requires_payment_method = _(
        "Intent created and requires a Payment Method to be attached."
    )
    requires_confirmation = _("Intent is ready to be confirmed.")
    requires_action = _("Payment Method require additional action, such as 3D secure.")
    processing = _("Required actions have been handled.")
    canceled = _(
        "Cancellation invalidates the intent for future confirmation and "
        "cannot be undone."
    )
    succeeded = _(
        "Setup was successful and the payment method is optimized for future payments."
    )


class PayoutFailureCode(Enum):
    """
    Payout failure error codes.

    https://stripe.com/docs/api#payout_failures
    """

    account_closed = _("Bank account has been closed.")
    account_frozen = _("Bank account has been frozen.")
    bank_account_restricted = _("Bank account has restrictions on payouts allowed.")
    bank_ownership_changed = _("Destination bank account has changed ownership.")
    could_not_process = _("Bank could not process payout.")
    debit_not_authorized = _("Debit transactions not approved on the bank account.")
    insufficient_funds = _("Stripe account has insufficient funds.")
    invalid_account_number = _("Invalid account number")
    invalid_currency = _("Bank account does not support currency.")
    no_account = _("Bank account could not be located.")
    unsupported_card = _("Card no longer supported.")


class PayoutMethod(Enum):
    standard = _("Standard")
    instant = _("Instant")


class PayoutStatus(Enum):
    paid = _("Paid")
    pending = _("Pending")
    in_transit = _("In transit")
    canceled = _("Canceled")
    failed = _("Failed")


class PayoutType(Enum):
    bank_account = _("Bank account")
    card = _("Card")


class PaymentIntentCancellationReason(Enum):
    # see also SetupIntentCancellationReason
    # User provided reasons:
    duplicate = _("Duplicate")
    fraudulent = _("Fraudulent")
    abandoned = _("Abandoned")
    requested_by_customer = _("Requested by Customer")
    # Reasons generated by Stripe internally
    failed_invoice = _("Failed invoice")
    void_invoice = _("Void invoice")
    automatic = _("Automatic")


class PlanAggregateUsage(Enum):
    last_during_period = _("Last during period")
    last_ever = _("Last ever")
    max = _("Max")
    sum = _("Sum")


class PlanBillingScheme(Enum):
    per_unit = _("Per unit")
    tiered = _("Tiered")


class PlanInterval(Enum):
    day = _("Day")
    week = _("Week")
    month = _("Month")
    year = _("Year")


class PlanUsageType(Enum):
    metered = _("Metered")
    licensed = _("Licensed")


class PlanTiersMode(Enum):
    graduated = _("Graduated")
    volume = _("Volume-based")


class ProductType(Enum):
    good = _("Good")
    service = _("Service")


class SetupIntentCancellationReason(Enum):
    # see also PaymentIntentCancellationReason
    abandoned = _("Abandoned")
    requested_by_customer = _("Requested by Customer")
    duplicate = _("Duplicate")


class ScheduledQueryRunStatus(Enum):
    canceled = _("Canceled")
    failed = _("Failed")
    timed_out = _("Timed out")


class SourceFlow(Enum):
    redirect = _("Redirect")
    receiver = _("Receiver")
    code_verification = _("Code verification")
    none = _("None")


class SourceStatus(Enum):
    canceled = _("Canceled")
    chargeable = _("Chargeable")
    consumed = _("Consumed")
    failed = _("Failed")
    pending = _("Pending")


class SourceType(Enum):
    ach_credit_transfer = _("ACH Credit Transfer")
    ach_debit = _("ACH Debit")
    alipay = _("Alipay")
    bancontact = _("Bancontact")
    bitcoin = _("Bitcoin")
    card = _("Card")
    card_present = _("Card present")
    eps = _("EPS")
    giropay = _("Giropay")
    ideal = _("iDEAL")
    p24 = _("P24")
    paper_check = _("Paper check")
    sepa_debit = _("SEPA Direct Debit")
    sepa_credit_transfer = _("SEPA credit transfer")
    sofort = _("SOFORT")
    three_d_secure = _("3D Secure")


class LegacySourceType(Enum):
    card = _("Card")
    bank_account = _("Bank account")
    bitcoin_receiver = _("Bitcoin receiver")
    alipay_account = _("Alipay account")


class RefundFailureReason(Enum):
    lost_or_stolen_card = _("Lost or stolen card")
    expired_or_canceled_card = _("Expired or canceled card")
    unknown = _("Unknown")


class RefundReason(Enum):
    duplicate = _("Duplicate charge")
    fraudulent = _("Fraudulent")
    requested_by_customer = _("Requested by customer")
    expired_uncaptured_charge = _("Expired uncaptured charge")


class RefundStatus(Enum):
    pending = _("Pending")
    succeeded = _("Succeeded")
    failed = _("Failed")
    canceled = _("Canceled")


class SessionBillingAddressCollection(Enum):
    auto = _("Auto")
    required = _("Required")


class SessionMode(Enum):
    payment = _("Payment")
    setup = _("Setup")
    subscription = _("Subscription")


class SourceUsage(Enum):
    reusable = _("Reusable")
    single_use = _("Single-use")


class SourceCodeVerificationStatus(Enum):
    pending = _("Pending")
    succeeded = _("Succeeded")
    failed = _("Failed")


class SourceRedirectFailureReason(Enum):
    user_abort = _("User-aborted")
    declined = _("Declined")
    processing_error = _("Processing error")


class SourceRedirectStatus(Enum):
    pending = _("Pending")
    succeeded = _("Succeeded")
    not_required = _("Not required")
    failed = _("Failed")


class SubmitTypeStatus(Enum):
    auto = _("Auto")
    book = _("Book")
    donate = _("donate")
    pay = _("pay")


class SubscriptionStatus(Enum):
    incomplete = _("Incomplete")
    incomplete_expired = _("Incomplete Expired")
    trialing = _("Trialing")
    active = _("Active")
    past_due = _("Past due")
    canceled = _("Canceled")
    unpaid = _("Unpaid")


class DjstripePaymentMethodType(Enum):
    """
    A djstripe-specific enum for the DjStripePaymentMethod model.
    """

    card = _("Card")
    bank_account = _("Bank account")
    source = _("Source")



"""
signals are sent for each event Stripe sends to the app

Stripe docs for Webhooks: https://stripe.com/docs/webhooks
"""
from django.db.models.signals import pre_delete
from django.dispatch import Signal, receiver

from . import settings as djstripe_settings

webhook_processing_error = Signal(providing_args=["data", "exception"])

# A signal for each Event type. See https://stripe.com/docs/api/events/types

WEBHOOK_SIGNALS = dict(
    [
        (hook, Signal(providing_args=["event"]))
        for hook in [
            # Update this by copy-pasting the "enabled_events" enum values from
            # https://raw.githubusercontent.com/stripe/openapi/master/openapi/spec3.json
            "account.application.authorized",
            "account.application.deauthorized",
            "account.external_account.created",
            "account.external_account.deleted",
            "account.external_account.updated",
            "account.updated",
            "application_fee.created",
            "application_fee.refund.updated",
            "application_fee.refunded",
            "balance.available",
            "capability.updated",
            "charge.captured",
            "charge.dispute.closed",
            "charge.dispute.created",
            "charge.dispute.funds_reinstated",
            "charge.dispute.funds_withdrawn",
            "charge.dispute.updated",
            "charge.expired",
            "charge.failed",
            "charge.pending",
            "charge.refund.updated",
            "charge.refunded",
            "charge.succeeded",
            "charge.updated",
            "checkout.session.completed",
            "coupon.created",
            "coupon.deleted",
            "coupon.updated",
            "credit_note.created",
            "credit_note.updated",
            "credit_note.voided",
            "customer.created",
            "customer.deleted",
            "customer.discount.created",
            "customer.discount.deleted",
            "customer.discount.updated",
            "customer.source.created",
            "customer.source.deleted",
            "customer.source.expiring",
            "customer.source.updated",
            "customer.subscription.created",
            "customer.subscription.deleted",
            "customer.subscription.trial_will_end",
            "customer.subscription.updated",
            "customer.tax_id.created",
            "customer.tax_id.deleted",
            "customer.tax_id.updated",
            "customer.updated",
            "file.created",
            "invoice.created",
            "invoice.deleted",
            "invoice.finalized",
            "invoice.marked_uncollectible",
            "invoice.payment_action_required",
            "invoice.payment_failed",
            "invoice.payment_succeeded",
            "invoice.sent",
            "invoice.upcoming",
            "invoice.updated",
            "invoice.voided",
            "invoiceitem.created",
            "invoiceitem.deleted",
            "invoiceitem.updated",
            "issuing_authorization.created",
            "issuing_authorization.request",
            "issuing_authorization.updated",
            "issuing_card.created",
            "issuing_card.updated",
            "issuing_cardholder.created",
            "issuing_cardholder.updated",
            "issuing_dispute.created",
            "issuing_dispute.updated",
            "issuing_settlement.created",
            "issuing_settlement.updated",
            "issuing_transaction.created",
            "issuing_transaction.updated",
            "mandate.updated",
            "order.created",
            "order.payment_failed",
            "order.payment_succeeded",
            "order.updated",
            "order_return.created",
            "payment_intent.amount_capturable_updated",
            "payment_intent.canceled",
            "payment_intent.created",
            "payment_intent.payment_failed",
            "payment_intent.succeeded",
            "payment_method.attached",
            "payment_method.card_automatically_updated",
            "payment_method.detached",
            "payment_method.updated",
            "payout.canceled",
            "payout.created",
            "payout.failed",
            "payout.paid",
            "payout.updated",
            "person.created",
            "person.deleted",
            "person.updated",
            "plan.created",
            "plan.deleted",
            "plan.updated",
            "product.created",
            "product.deleted",
            "product.updated",
            "radar.early_fraud_warning.created",
            "radar.early_fraud_warning.updated",
            "recipient.created",
            "recipient.deleted",
            "recipient.updated",
            "reporting.report_run.failed",
            "reporting.report_run.succeeded",
            "reporting.report_type.updated",
            "review.closed",
            "review.opened",
            "setup_intent.canceled",
            "setup_intent.created",
            "setup_intent.setup_failed",
            "setup_intent.succeeded",
            "sigma.scheduled_query_run.created",
            "sku.created",
            "sku.deleted",
            "sku.updated",
            "source.canceled",
            "source.chargeable",
            "source.failed",
            "source.mandate_notification",
            "source.refund_attributes_required",
            "source.transaction.created",
            "source.transaction.updated",
            "subscription_schedule.aborted",
            "subscription_schedule.canceled",
            "subscription_schedule.completed",
            "subscription_schedule.created",
            "subscription_schedule.expiring",
            "subscription_schedule.released",
            "subscription_schedule.updated",
            "tax_rate.created",
            "tax_rate.updated",
            "topup.canceled",
            "topup.created",
            "topup.failed",
            "topup.reversed",
            "topup.succeeded",
            "transfer.created",
            "transfer.failed",
            "transfer.paid",
            "transfer.reversed",
            "transfer.updated",
            # deprecated (no longer in events_types list) - TODO can be deleted?
            "checkout_beta.session_succeeded",
            "issuer_fraud_record.created",
            "payment_intent.requires_capture",
            "subscription_schedule.canceled",
            "subscription_schedule.completed",
            "subscription_schedule.created",
            "subscription_schedule.released",
            "subscription_schedule.updated",
            # special case? - TODO can be deleted?
            "ping",
        ]
    ]
)


@receiver(pre_delete, sender=djstripe_settings.get_subscriber_model_string())
def on_delete_subscriber_purge_customer(instance=None, **kwargs):
    """ Purge associated customers when the subscriber is deleted. """
    for customer in instance.djstripe_customers.all():
        customer.purge()



import stripe
from django.db import models

from .. import enums
from ..fields import JSONField, StripeEnumField
from .base import StripeModel


class Session(StripeModel):
    """
    A Checkout Session represents your customer's session as they pay
    for one-time purchases or subscriptions through Checkout.
    """

    stripe_class = stripe.checkout.Session

    billing_address_collection = StripeEnumField(
        enum=enums.SessionBillingAddressCollection,
        blank=True,
        help_text=(
            "The value (auto or required) for whether Checkout"
            "collected the customer’s billing address."
        ),
    )
    cancel_url = models.TextField(
        max_length=5000,
        blank=True,
        help_text=(
            "The URL the customer will be directed to if they"
            "decide to cancel payment and return to your website."
        ),
    )
    client_reference_id = models.TextField(
        max_length=5000,
        blank=True,
        help_text=(
            "A unique string to reference the Checkout Session."
            "This can be a customer ID, a cart ID, or similar, and"
            "can be used to reconcile the session with your internal systems."
        ),
    )
    customer = models.ForeignKey(
        "Customer",
        null=True,
        on_delete=models.SET_NULL,
        help_text=("Customer this Checkout is for if one exists."),
    )
    customer_email = models.CharField(
        max_length=255,
        blank=True,
        help_text=(
            "If provided, this value will be used when the Customer object is created."
        ),
    )
    display_items = JSONField(
        null=True,
        blank=True,
        help_text=("The line items, plans, or SKUs purchased by the customer."),
    )
    locale = models.CharField(
        max_length=255,
        blank=True,
        help_text=(
            "The IETF language tag of the locale Checkout is displayed in."
            "If blank or auto, the browser’s locale is used."
        ),
    )
    mode = StripeEnumField(
        enum=enums.SessionMode,
        blank=True,
        help_text="The mode of the Checkout Session, "
        "one of payment, setup, or subscription.",
    )
    payment_intent = models.ForeignKey(
        "PaymentIntent",
        null=True,
        on_delete=models.SET_NULL,
        help_text=("PaymentIntent created if SKUs or line items were provided."),
    )
    payment_method_types = JSONField(
        help_text="The list of payment method types (e.g. card) that this "
        "Checkout Session is allowed to accept."
    )
    submit_type = StripeEnumField(
        enum=enums.SubmitTypeStatus,
        blank=True,
        help_text="Describes the type of transaction being performed by Checkout"
        "in order to customize relevant text on the page, such as the submit button.",
    )
    subscription = models.ForeignKey(
        "Subscription",
        null=True,
        on_delete=models.SET_NULL,
        help_text=("Subscription created if one or more plans were provided."),
    )
    success_url = models.TextField(
        max_length=5000,
        blank=True,
        help_text=(
            "The URL the customer will be directed to after the payment or subscription"
            "creation is successful."
        ),
    )



#!/usr/bin/env python
import os
import sys

from setuptools import setup

if sys.argv[-1] == "publish":
    os.system("python setup.py bdist_wheel upload --sign")
    sys.exit()


readme = open("README.rst").read()
history = open("HISTORY.rst").read().replace(".. :changelog:", "")


setup(long_description=readme + "\n\n" + history)



"""
dj-stripe - Views related to the djstripe app.
"""
import logging

from django.http import HttpResponse, HttpResponseBadRequest
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from .models import WebhookEventTrigger

logger = logging.getLogger(__name__)


@method_decorator(csrf_exempt, name="dispatch")
class ProcessWebhookView(View):
    """
    A Stripe Webhook handler view.

    This will create a WebhookEventTrigger instance, verify it,
    then attempt to process it.

    If the webhook cannot be verified, returns HTTP 400.

    If an exception happens during processing, returns HTTP 500.
    """

    def post(self, request):
        if "HTTP_STRIPE_SIGNATURE" not in request.META:
            # Do not even attempt to process/store the event if there is
            # no signature in the headers so we avoid overfilling the db.
            return HttpResponseBadRequest()

        trigger = WebhookEventTrigger.from_request(request)

        if trigger.is_test_event:
            # Since we don't do signature verification, we have to skip trigger.valid
            return HttpResponse("Test webhook successfully received!")

        if not trigger.valid:
            # Webhook Event did not validate, return 400
            return HttpResponseBadRequest()

        return HttpResponse(str(trigger.id))



import stripe
from django.db import models

from .. import enums
from ..fields import JSONField, StripeDateTimeField, StripeEnumField
from .base import StripeModel


class ScheduledQueryRun(StripeModel):
    """
    Stripe documentation: https://stripe.com/docs/api#scheduled_queries
    """

    stripe_class = stripe.sigma.ScheduledQueryRun

    data_load_time = StripeDateTimeField(
        help_text="When the query was run, Sigma contained a snapshot of your "
        "Stripe data at this time."
    )
    error = JSONField(
        null=True,
        blank=True,
        help_text="If the query run was not succeesful, contains information "
        "about the failure.",
    )
    file = models.ForeignKey(
        "FileUpload",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        help_text="The file object representing the results of the query.",
    )
    result_available_until = StripeDateTimeField(
        help_text="Time at which the result expires and is no longer available "
        "for download."
    )
    sql = models.TextField(max_length=5000, help_text="SQL for the query.")
    status = StripeEnumField(
        enum=enums.ScheduledQueryRunStatus, help_text="The query's execution status."
    )
    title = models.TextField(max_length=5000, help_text="Title of the query.")
